/*
 * Copyright (c) 1992-2001 David Gay and
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include "regions.h"

/* Module: regions
   Date: 14 Feb 92
   Purpose: Fast & easy memory allocator for the mudlle compiler
     Based on the concept of blocks: you allocate memory from a block,
       and can free the whole block at one go.
       Individual deallocations are not possible.
*/

#define ALIGN(x, n) (((x) + ((n) - 1)) & ~((n) - 1))

#define BLOCK_SIZE 8172		/* Should probably be chosen w/ respect to
				   malloc implementation */
#define BIG_SIZE 16364

typedef double max_aligned_type;

struct region
{
#ifndef USEMALLOC
  char *pos;			/* Where next to allocate from */
  char *end;			/* End of block */
#endif
  struct region *previous;
  max_aligned_type data[1];
};

nomem_handler nomem_h;

#ifdef BWGC
region newregion(void)
{
  return NULL;
}

void deleteregion(region b)
/* Effect: Free all memory allocated in region b.
*/
{
}

void deleteregion_ptr(region *r)
/* Effect: Free all memory allocated in region *r and sets *r to NULL.
*/
{
}

static void *allocate(region b, unsigned long size)
/* Effects: Allocates size bytes from block b. The result is aligned
     correctly for all types.
   Returns: A pointer to the start of the block.
   Note: In this implementation, 12 + average(size)/2 bytes will be wasted
     for every BLOCK_SIZE bytes allocated.
*/
{
  return GC_malloc(size);
}
#else

static struct region *make_block(int size)
{
  struct region *newp = malloc(offsetof(struct region, data) + size);

  if (!newp)
    return NULL;

#ifndef USEMALLOC
  newp->pos = (char *)newp->data;
  newp->end = newp->pos + size;
#endif
  newp->previous = 0;

  return newp;
}

region newregion(void)
/* Return: A new region from which to allocate some memory.
*/
{
  region newp = malloc(sizeof *newp);

  if (!newp)
    return nomem_h();

  *newp = make_block(BLOCK_SIZE);
  if (!*newp)
    {
      free(newp);
      return nomem_h();
    }
      
  return newp;
}

void deleteregion(region b)
/* Effect: Free all memory allocated in region b.
*/
{
  struct region *blk = *b;

  while (blk)
    {
      struct region *prev = blk->previous;

      free(blk);
      blk = prev;
    }
}

void deleteregion_ptr(region *r)
/* Effect: Free all memory allocated in region *r and sets *r to NULL.
*/
{
  deleteregion(*r);
  *r = NULL;
}

static void *allocate(region b, unsigned long size)
/* Effects: Allocates size bytes from block b. The result is aligned
     correctly for all types.
   Returns: A pointer to the start of the block.
   Note: In this implementation, 12 + average(size)/2 bytes will be wasted
     for every BLOCK_SIZE bytes allocated.
*/
{
#ifdef USEMALLOC
  struct region *old = *b, *new = make_block(size);

  new->previous = old;
  *b = new;
  return &new->data;
#else
  struct region *blk = *b;
  void *result;

  size = ALIGN(size, sizeof(max_aligned_type));

  result = blk->pos;
  if ((blk->pos += size) >= blk->end)
    {
      /* Block full, get new one */
      int bsize;
      struct region *newp;

      if (size < BLOCK_SIZE / 4)
	bsize = BLOCK_SIZE;
      else if (size >= BIG_SIZE)
	bsize = size;
      else
	bsize = size * 4;
      newp = make_block(bsize);

      if (!newp)
	return nomem_h();

      newp->previous = blk;
      *b = newp;
      result = newp->pos;
      newp->pos += size;
    }
  return result;
#endif
}
#endif

void *int_ralloc(region r, size_t size)
{
  void *newo = allocate(r, size);

  if (newo)
    memset(newo, 0, size);

  return newo;
}

void *int_rarrayalloc(region r, size_t n, size_t size)
{
  return int_ralloc(r, n * size);
}

char *rstralloc(region r, size_t size)
{
  return allocate(r, size);
}

char *rstrdup(region r, const char *s)
{
  int l = strlen(s);
  char *news = allocate(r, l + 1);

  if (news)
    memcpy(news, s, l + 1);

  return news;
}

nomem_handler set_nomem_handler(nomem_handler newhandler)
{
  nomem_handler oldh = nomem_h;

  nomem_h = newhandler;

  return oldh;
}

int region_main(int argc, char **argv, char **envp);

int main(int argc, char **argv, char **envp)
{
  return region_main(argc, argv, envp);
}


